from itertools import count
import numpy as np
from stats_utils import Bar, AverageMeter
import time
import torch.multiprocessing as mp


def init_processes(_counter):
    global counter
    counter = _counter


class Arena():
    """
    An Arena class where any 2 agents can be pit against each other.
    """
    def __init__(self, player1, player2, game, numPs, display=None):
        """
        Input:
            player 1,2: two MCTS classes, one with old NN, onw with new NN
            game: Game object
            numPs: the number of processes to execute at the same time
            display: a function that takes board as input and prints it (not used)
        """
        self.player1 = player1
        self.player2 = player2
        self.game = game
        self.numPs = numPs
        self.display = display
        self.tempThreshold = 5

    def playGame(self, eps, num, total_beg, verbose=False):
        """
        Executes one episode of a game.

        Returns:
            either
                winner: player who won the game (1 if player1, -1 if player2)
            or
                draw result returned from the game that is neither 1, -1, nor 0.
        """
        global counter
        eps_beg = time.time()
        if eps % 2:
            players = [self.player2, None, self.player1]
        else:
            players = [self.player1, None, self.player2]
        players[0].args.numMCTSSims = 25
        players[2].args.numMCTSSims = 25
        curPlayer = 1
        board = self.game.getInitBoard()
        it = 0
        while self.game.getGameEnded(board, curPlayer)==0:
            temp = int(it < self.tempThreshold)
            it+=1
            if verbose:
                assert(self.display)
                print("Turn ", str(it), "Player ", str(curPlayer))
                self.display(board)
            pi = players[curPlayer+1].getActionProb(self.game.getCanonicalBoard(board, curPlayer), temp)
            action = np.random.choice(len(pi), p=pi)

            valids = self.game.getValidMoves(self.game.getCanonicalBoard(board, curPlayer),1)

            if valids[action]==0:
                print(action)
                assert valids[action] >0
            board, curPlayer = self.game.getNextState(board, curPlayer, action)
        if verbose:
            assert(self.display)
            print("Game over: Turn ", str(it), "Result ", str(self.game.getGameEnded(board, 1)))
            self.display(board)

        eps_end = total_end = time.time()
        with counter.get_lock():
            counter.value += 1
            print("{}/{} | Eps Time: {:.3f}s | Total: {:.3f}s".format(counter.value, num, eps_end-eps_beg, total_end-total_beg))
        return self.game.getGameEnded(board, curPlayer) * ((-1)**(eps % 2 == 0))

    def playGames(self, num, verbose=False):
        """
        Plays num games in which player1 starts num/2 games and player2 starts
        num/2 games.

        Returns:
            oneWon: games won by player1
            twoWon: games won by player2
            draws:  games won by nobody
        """
        counter = mp.Value('i', 0)

        oneWon = 0
        twoWon = 0
        total_beg = time.time()
        with mp.Pool(processes=self.numPs, initializer=init_processes, initargs=(counter,)) as p:
            self.player1.clear()
            self.player2.clear()
            results = p.starmap(self.playGame, [(eps, num, total_beg) for eps in range(1, num+1)])
        oneWon += results.count(1)
        twoWon += results.count(-1)
        draws = num - oneWon - twoWon

        return oneWon, twoWon, draws
