import random, time, os
from SelfPlay import SelfPlay
from Arena import Arena
from MCTS import MCTS
import numpy as np
from pickle import Pickler, Unpickler
from random import shuffle


class Coach:

    def __init__(self, game, nnet, args):
        self.game = game        # NogoGame instance
        self.nnet = nnet        # NNetWrapper instance
        self.pnet = self.nnet.__class__(self.game)  # the previous network
        self.args = args
        self.mcts = MCTS(self.game, self.nnet, self.args)
        self.trainExamplesHistory = []  # history of examples from args.numItersForTrainExamplesHistory latest iterations
        self.skipFirstSelfPlay = False

    def learn(self):
        """
        Performs numIters iterations with numEps episodes of self-play in each
        iteration. After every iteration, it retrains neural network with
        examples in trainExamples (which has a maximum length of maxlenofQueue).
        It then pits the new neural network against the old one and accepts it
        only if it wins >= updateThreshold fraction of games.
        """
        for i in range(1, self.args.numIters+1):
            print("------ITER " + str(i) + "------")
            if not self.skipFirstSelfPlay or i > 1:
                sp = SelfPlay(self.mcts, self.game, self.args)
                iterationTrainExamples = sp.playGames()

                assert len(iterationTrainExamples) <= self.args.maxlenOfQueue
                self.trainExamplesHistory.append(iterationTrainExamples)

            if len(self.trainExamplesHistory) > self.args.numItersForTrainExamplesHistory:
                print("len(trainExamplesHistory) =", len(self.trainExamplesHistory), " => remove the oldest trainExamples")
                self.trainExamplesHistory.pop(0)
            self.saveTrainExamples(i-1)     # backup history to a file

            # shuffle examples before training
            trainExamples = []
            for e in self.trainExamplesHistory:
                trainExamples.extend(e)
            shuffle(trainExamples)

            # train new network, keeping a copy of the old one
            self.nnet.save_checkpoint(folder=self.args.checkpoint, filename="temp.pth.tar")
            self.pnet.load_checkpoint(folder=self.args.checkpoint, filename="temp.pth.tar")
            pmcts = MCTS(self.game, self.pnet, self.args)

            self.nnet.train(trainExamples)
            nmcts = MCTS(self.game, self.nnet, self.args)

            print("PITTING AGAINST PREVIOUS VERSION")
            arena = Arena(pmcts, nmcts, self.game, self.args.numPs)
            pwins, nwins, draws = arena.playGames(self.args.arenaCompare)

            print("NEW/PREV WINS : %d / %d ; DRAWS : %d" % (nwins, pwins, draws))
            if (pwins + nwins) == 0 or float(nwins)/(pwins+nwins) < self.args.updateThreshold:
                print("REJECTING NEW MODEL")
                self.nnet.load_checkpoint(folder=self.args.checkpoint, filename="temp.pth.tar")
            else:
                print("ACCEPTING NEW MODEL")
                self.nnet.save_checkpoint(folder=self.args.checkpoint, filename=self.getCheckpointFile(i))
                self.nnet.save_checkpoint(folder=self.args.checkpoint, filename="best.pth.tar")

    def getCheckpointFile(self, iteration):
        return "checkpoint_" + str(iteration) + ".pth.tar"

    def saveTrainExamples(self, iteration):
        folder = self.args.checkpoint
        if not os.path.exists(folder):
            os.makedirs(folder)
        filename = os.path.join(folder, self.getCheckpointFile(iteration) + ".examples")
        with open(filename, "wb+") as f:
            Pickler(f).dump(self.trainExamplesHistory)

    def loadTrainExamples(self):
        file = input("Enter full path to trainExamples, [return] to skip: ")
        if not file:
            return
        if not os.path.isfile(file):
            print("File with trainExamples not found.\n")
            self.loadTrainExamples()
            return
        with open(file, "rb") as f:
            self.trainExamplesHistory = Unpickler(f).load()
        self.skipFirstSelfPlay = True
        return
