import random, time
import numpy as np
import torch.multiprocessing as mp


def init_processes(_counter):
    global counter
    counter = _counter


class SelfPlay:
    """
    A SelfPlay class where the agent play against itself.
    """
    def __init__(self, mcts, game, args):
        self.mcts = mcts
        self.game = game
        self.args = args
    
    def playGame(self, total_beg):
        """
        This function executes one episode of self-play, starting with player 1.
        As the game is played, each turn is added as a training example to
        trainExamples. The game is played till the game ends. After the game
        ends, the outcome of the game is used to assign values to each example
        in trainExamples.

        It uses a temp=1 if episodeStep < tempThreshold, and thereafter
        uses temp=0.

        Returns:
            trainExamples: a list of examples of the form (canonicalBoard, currPlayer, pi,v)
                           pi is the MCTS informed policy vector, v is +1 if
                           the player eventually won the game, else -1.
        """
        global counter
        eps_beg = time.time()
        trainExamples = []
        canonicalBoard = self.game.getInitBoard()       # reset game
        toPlay = 1
        episodeStep = 0

        while True:
            episodeStep += 1
            # when temperature is high (first tempThreshold-1 moves), using probabilistic policy
            # when temperature is low (all other moves), using the most visited child node
            temp = int(episodeStep < self.args.tempThreshold)

            pi = self.mcts.getActionProb(canonicalBoard, temp=temp)
            syms = self.game.getSymmetries(canonicalBoard, pi)
            for b, p in syms:
                trainExamples.append([b, toPlay, p, None])
            
            # dirichlet noise
            pi = np.array(pi)
            idx = np.nonzero(pi)[0]
            alpha = [self.args.dirichlet_alpha] * len(idx)
            dirichlet_sample = np.random.dirichlet(alpha)
            noise = np.zeros(self.game.getActionSize())
            np.add.at(noise, idx, dirichlet_sample)
            pi_noise = (1-self.args.eps) * pi + self.args.eps * noise

            action = np.random.choice(len(pi_noise), p=pi_noise)
            canonicalBoard, toPlay = self.game.getNextState(canonicalBoard, toPlay, action)
            r = self.game.getGameEnded(canonicalBoard, toPlay)

            if r != 0:  # game ended
                l = [(x[0], x[2], (-1)**(x[1]==toPlay)) for x in trainExamples]
                sample_size = int(self.args.maxlenOfQueue / self.args.numEps)
                if sample_size < len(l):
                    l = random.sample(l, sample_size)
                total_end = eps_end = time.time()
                with counter.get_lock():
                    counter.value += 1
                    print("Self Play {}/{} | Eps Time: {:.3f}s | Total: {:.3f}s".format(
                        counter.value, self.args.numEps, eps_end-eps_beg, total_end-total_beg))
                return l
    
    def playGames(self):
        """
        Plays self.args.numEps self-play games.

        Returns:
            training examples sampled from self-play games
        """
        iterationTrainExamples = []
        counter = mp.Value("i", 0)

        total_beg = time.time()
        with mp.Pool(processes=self.args.numPs, initializer=init_processes, initargs=(counter,)) as p:
            self.mcts.clear()
            examples = p.map(self.playGame, [total_beg]*self.args.numEps)
        
        for e in examples:
            iterationTrainExamples += e
        
        return iterationTrainExamples
        
