import sys
sys.path.append('..')
from Game import Game
from .board import GoBoard as Board
from .board_util import GoBoardUtil, EMPTY, BLACK, WHITE, coord_to_point
import numpy as np

class NogoGame(Game):
    square_content = {
        -1: "w",
        +0: "-",
        +1: "b"
    }

    def __init__(self, n):
        self.n = n      # board size
        self.board = Board(n)
        self.root = None

    def getInitBoard(self):
        self.board.reset(self.n)
        return self.get_pieces()

    def getBoardSize(self):
        return self.n, self.n

    def getActionSize(self):
        return self.n * self.n

    def getNextState(self, board, player, action):
        """
        Inputs:
            board: current board (unused)
            player: current player (unused)
            action: a cell on board
        Returns:
            nextBoard: canonical board after applying the action
            nextPlayer: canonical player after applying the action
        """
        point = self.cell_to_point(action)
        assert self.board.play_move(point, self.board.current_player)
        return self.getCanonicalBoard(board, player), self.getCanonicalPlayer()

    def getValidMoves(self, board, player, search=False):
        # return all empty cells on board during MCTS (search=True)
        # otherwise, return only valid cells (search=False)
        # a binary vector (0 or 1) of size self.n**2 is returned
        points = self.board.get_empty_points()
        if search:
            legal = points
        else:
            legal = []
            cboard = self.board.copy()
            for point in points:
                if cboard.play_move(point, self.board.current_player):
                    cboard.board[point] = EMPTY     # undo move
                    legal.append(point)
        valids = [0] * self.getActionSize()
        for point in legal:
            valids[self.point_to_cell(point)] = 1
        return np.array(valids)

    def getGameEnded(self, board, player, search=False):
        # return 0 during MCTS (search=True)
        # otherwise, return not ended (0), black wins (+1), white wins (-1)
        if search:
            return 0
        legal_moves = self.getValidMoves(board, player, search=False)
        if legal_moves.max() == 1:
            return 0
        else:
            return -self.getCanonicalPlayer()

    def getSymmetries(self, board, pi):
        assert (len(pi) == self.n*self.n)
        pi_board = np.flipud(np.reshape(pi, (self.n, self.n)))
        l = []

        for i in range(1, 5):
            newB = np.rot90(board, i)
            newPi = np.rot90(pi_board, i)
            newB_mirror = np.fliplr(newB)
            newPi_mirror = np.fliplr(newPi)
            l += [(newB, list(np.flipud(newPi).ravel())), (newB_mirror, list(np.flipud(newPi_mirror).ravel()))]
        return l

    def stringRepresentation(self, board):
        """
        Inputs:
            board: 2d numpy array
        Returns:
            boardBytes: board of type <class 'bytes'>
        """
        return board.tobytes()

    def stringRepresentationReadable(self, board):
        board_s = "".join(self.square_content[square] for row in board for square in row)
        return board_s

    def get_pieces(self):
        # return numpy 2d array with black=1, white=-1, empty=0
        pieces = GoBoardUtil.get_twoD_board(self.board)
        whites = (pieces == 2).astype(int)
        whites = np.multiply(-3, whites)
        pieces = np.add(pieces, whites)
        return pieces

    def getCanonicalBoard(self, board, player):
        """
        Inputs:
            board: current board (not used)
            player: current player (not used)
        Returns:
            canonicalBoard: canonical form of the board
        """
        return self.getCanonicalPlayer() * self.get_pieces()

    def getCanonicalPlayer(self):
        if self.board.current_player == WHITE:
            return -1
        if self.board.current_player == BLACK:
            return 1
        return None

    def cell_to_point(self, cell):
        x, y = divmod(cell, self.n)
        return coord_to_point(x+1, y+1, self.n)

    def point_to_cell(self, point):
        row, col = self.board._point_to_coord(point)
        return (row-1) * self.n + (col-1)

    def beginSearch(self):
        self.root = self.board.copy()

    def inSearch(self):
        self.board = self.root.copy()

    def endSearch(self):
        self.board = self.root.copy()
        self.root = None
