
from nogo.gtp_connection import GtpConnection
from nogo.board_util import GoBoardUtil, BLACK, WHITE
from nogo.board import GoBoard as GoBoard
from nogo.NogoGame import NogoGame
from MCTS import MCTS
from nogo.pytorch.NNet import NNetWrapper as NNet
from utils import *
import numpy as np


class NoGo:
    def __init__(self, size):
        self.name = "AlphaZero"
        self.version = 3.0
        self.size = size
        self.args = dotdict({'numMCTSSims': 1000, 'cpuct': 1.0})
        self.game = NogoGame(self.size)
        self.nn = NNet(self.game)
        self.nn.load_checkpoint("./temp", f"best{size}x{size}.pth.tar")
        self.mcts = MCTS(self.game, self.nn, self.args)

        self.clear_tree = True              # clear tree between state
        self.numSimsPerValidMove = 100      # simulation budget per valid move
        self.noise = True                   # noise at the beginning; play probabilistic pi
        self.play_nn = False                # play by NN only; no search

    def get_policy(self, board):
        """
        Inputs:
            board - GoBaord
        Returns:
            move - point on board
        """
        self.mcts.game.board = board    # update board
        if self.clear_tree:
            self.mcts.clear()               # clear tree

        canonicalBoard = self.game.getCanonicalBoard(self.game.board, 1)

        # simulation budget
        valids = self.game.getValidMoves(canonicalBoard, 1, search=False)
        if self.numSimsPerValidMove:
            self.mcts.args.numMCTSSims = np.count_nonzero(valids) * self.numSimsPerValidMove

        # noise at the beginning of game
        if self.noise:
            length = np.count_nonzero(canonicalBoard)
            temp = int(length < 4)
        else:
            temp = 0

        # policy
        if self.play_nn:
            pi, v = self.nn.predict(canonicalBoard)
            pi = pi * valids
            pi_sum = pi.sum()
            if pi_sum == 0.0:
                return None
            pi = pi / pi_sum
            if not temp:
                best_a = np.argmax(pi)
                pi = [0] * self.game.getActionSize()
                pi[best_a] = 1
        else:
            pi = self.mcts.getActionProb(canonicalBoard, temp=temp)
        
        return pi
    
    def get_move(self, board):
        pi = self.get_policy(board)
        if pi is None:
            return None
        cell = np.random.choice(len(pi), p=pi)
        point = self.game.cell_to_point(cell)
        return point

    def load_model(self, size, path):
        self.size = size
        self.game = NogoGame(self.size)
        self.nn = NNet(self.game)
        self.nn.load_checkpoint("", path)
        self.mcts = MCTS(self.game, self.nn, self.args)
        board = GoBoard(size)
        return board


def run():
    """
    start the gtp connection and wait for commands.
    """
    size = 7
    board = GoBoard(size)
    con = GtpConnection(NoGo(size), board)
    con.start_connection()


if __name__ == "__main__":
    run()
